from typing import Union, List, Optional, Dict

import pandas as pd
from datetime import datetime
from deprecated import deprecated
from requests.exceptions import HTTPError
from snowflake import connector as snowflake
from snowflake.connector import SnowflakeConnection
from snowflake.connector.pandas_tools import write_pandas
import yaml

from pyrasgo.connection import Connection
from pyrasgo.enums import Granularity, ModelType
from pyrasgo.feature import Feature, FeatureList
from pyrasgo.model import Model
from pyrasgo.member import Member
from pyrasgo.monitoring import track_usage
from pyrasgo import schemas as api


class RasgoConnection(Connection):
    """
    Base connection object to handle interactions with the Rasgo API.

    Defaults to using the production Rasgo instance, which can be overwritten
    by specifying the `RASGO_DOMAIN` environment variable, eg:

    &> RASGO_DOMAIN=custom.rasgoml.com python
    >>> from pyrasgo import RasgoConnection
    >>> rasgo = Connection(api_key='not_a_key')
    >>> rasgo._hostname == 'custom.rasgoml.com'
    True

    """
    from pyrasgo.version import __version__

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @track_usage
    def create_model(self, name: str,
                     type: Union[str, ModelType],
                     granularity: Union[str, Granularity]) -> Model:
        """
        Creates model within Rasgo within the account specified by the API key.
        :param name: Model name
        :param model_type: Type of model specified
        :param granularity: Granularity of the data.
        :return: Model object.
        """
        try:
            # If not enum, convert to enum first.
            model_type = type.name
        except AttributeError:
            model_type = ModelType(type)

        try:
            # If not enum, convert to enum first.
            granularity = granularity.name
        except AttributeError:
            granularity = Granularity(granularity)

        # TODO: API should really have _some_ validation for the schema of this post.
        response = self._post("/models", _json={"name": name,
                                                "type": model_type.value,
                                                # TODO: This post should only be `granularity` for future compatibility
                                                #       Coordinate with API development to relax domain specificity
                                                "timeSeriesGranularity": granularity.value})
        return Model(api_key=self._api_key, username=self._username,
                     user_id=self._user_id, org_id=self._org_id, api_object=response.json())

    @track_usage
    def add_feature_to(self, model: Model, feature: Feature):
        model.add_feature(feature)

    @track_usage
    def add_features_to(self, model: Model, features: FeatureList):
        model.add_features(features)

    @track_usage
    def generate_training_data_for(self, model: Model):
        raise NotImplementedError

    @track_usage
    def get_models(self) -> List[Model]:
        """
        Retrieves the list of models from Rasgo within the account specified by the API key.
        """
        response = self._get("/models", {"join": ["features", "author"],
                                         "filter": "isDeleted||$isnull"})
        return [Model(api_key=self._api_key, username=self._username,
                      user_id=self._user_id, org_id=self._org_id, api_object=entry) for entry in response.json()]

    @track_usage
    def get_model(self, model_id) -> Model:
        """
        Retrieves the specified model from Rasgo within the account specified by the API key.
        """
        # TODO: The API should not throw a 500 if it can't find a non-existent model, it should return a 404
        try:
            self._get(f"/models/{model_id}")
        except HTTPError:
            # This assumes they've followed a work flow and had their API key verified (happens on class __init__).
            raise LookupError(f"The model (id={model_id}) could not be found in your Rasgo account.")
        # TODO: The API should not throw a 500 if it can't join against a non-existent model, it should return a 404.
        response = self._get(f"/models/{model_id}", {"join": ["features", "author"]})
        return Model(api_key=self._api_key, username=self._username,
                     user_id=self._user_id, org_id=self._org_id, api_object=response.json())

    @track_usage
    def get_feature(self, feature_id) -> Feature:
        """
        Retrieves the specified feature from Rasgo within the account specified by the API key.
        """
        response = self._get(f"/features/{feature_id}", api_version=1)
        return Feature(api_key=self._api_key, username=self._username,
                       user_id=self._user_id, org_id=self._org_id,
                       api_object=response.json())

    @track_usage
    def get_features(self) -> FeatureList:
        """
        Retrieves the features from Rasgo within account specified by the API key.
        """
        return FeatureList(api_key=self._api_key, username=self._username,
                           user_id=self._user_id, org_id=self._org_id,
                           api_object=self._get("/features", api_version=1).json())

    @track_usage
    def get_features_for(self, model: Model) -> FeatureList:
        raise NotImplementedError()
        # response = self._get("/features",
        #                      params={'join': f""})
        #                    # params={'filter': f"||eq||DateTime"})
        # return FeatureList(response.json())

    @track_usage
    def get_feature_data(self, model_id: int,
                         filters: Optional[Dict[str, str]] = None,
                         limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs the pandas dataframe for the specified model.

        :param model_id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        model = self.get_model(model_id)

        conn, creds = self._snowflake_connection(model.get_author())

        table_metadata = model.snowflake_table_metadata(creds)
        query, values = self._make_select_statement(table_metadata, filters, limit)

        result_set = self._run_query(conn, query, values)
        return pd.DataFrame.from_records(iter(result_set), columns=[x[0] for x in result_set.description])

    @track_usage
    def get_feature_sets(self):
        """
        Retrieves the feature sets from Rasgo within account specified by the API key.
        """
        return self._get("/feature-sets").json()

    @track_usage
    def get_source_tables(self):
        '''
        Retrieves a list of Snowflake tables and views that are queryable as feature sources
        '''
        conn, creds = self._snowflake_connection(self.get_member())
        query = 'select * from RASGO_DATA_SOURCE_TABLES'
        result_set = self._run_query(conn, query, None)
        return pd.DataFrame.from_records(iter(result_set), columns=[x[0] for x in result_set.description])

    @track_usage
    def get_source_table(self, table_name, record_limit: int):
        '''
        Returns top n records from a Snowflake source table to a dataframe
        '''
        conn, creds = self._snowflake_connection(self.get_member())
        query = f'select * from {table_name} limit {record_limit}'
        result_set = self._run_query(conn, query, None)
        return pd.DataFrame.from_records(iter(result_set), columns=[x[0] for x in result_set.description])

    @track_usage
    def get_source_columns(self):
        '''
        Retrieves a list of columns in Snowflake tables and views that are queryable as feature sources
        '''
        conn, creds = self._snowflake_connection(self.get_member())
        query = 'select * from RASGO_DATA_SOURCE_COLUMNS'
        result_set = self._run_query(conn, query, None)
        return pd.DataFrame.from_records(iter(result_set), columns=[x[0] for x in result_set.description])

    @track_usage
    def get_data_sources(self):
        """
        Retrieves the data sources from Rasgo within account specified by the API key.
        """
        return self._get("/data-source").json()

    @track_usage
    def get_dimensionalities(self):
        """
        Retrieves the data sources from Rasgo within account specified by the API key.
        """
        return self._get("/dimensionalities").json()

    @track_usage
    def get_datasource(self, name):
        rows = self.find("data-source", {"name": name})
        if 0 == len(rows):
            return None
        else:
            return rows[0]

    def get_member(self):
        return Member(self.get_user())

    def get_user(self):
        return self._get("/users/me", api_version=1).json()

    @track_usage
    def get_user_id(self):
        """
        Gets the user id for the API key provided, or None if not available.
        NOTE: This is used for monitoring/logging purposes.
        """
        if self._user_id:
            return self._user_id
        else:
            try:
                user = self.get_user()
                self._user_id = user.get('id')
                return self._user_id
            except Exception:
                return None

    @track_usage
    def get_dimensionality(self, name):
        return self._get_item("dimensionalities", {"name": name}, True)

    @track_usage
    def create_column(self, name, data_type, feature_set_id, dimension_id):
        column = api.ColumnCreate(name=name, dataType=data_type,
                                  featureSet=api.FeatureSetBase(id=feature_set_id),
                                  dimensionality=api.Dimensionality(id=dimension_id))
        return self._post("/rcolumns", column.dict(exclude_unset=True)).json()

    @track_usage
    def update_column(self, column_id, name, data_type, feature_set_id, dimension_id):
        column = api.ColumnUpdate(id=column_id, 
                                  name=name, dataType=data_type,
                                  featureSet=api.FeatureSetBase(id=feature_set_id),
                                  dimensionality=api.Dimensionality(id=dimension_id))
        return self._patch(f"rcolumns/{column_id}", column.dict(exclude_unset=True)).json()

    @track_usage
    def create_datasource(self, org_id, name):
        data_source = api.DataSourceCreate(name=name,
                                           abbreviation=name[:10].lower(),
                                           organization=api.Organization(id=org_id))
        return self._post("/data-source", data_source.dict(exclude_unset=True)).json()

    @track_usage
    def create_dimensionality(self, org_id, name, dimension_type, granularity):
        dimensionality = api.DimensionalityCreate(name=name,
                                                  dimension_type=dimension_type,
                                                  granularity=granularity,
                                                  organization=api.Organization(id=org_id))
        return self._post("/dimensionalities", dimensionality.dict(exclude_unset=True)).json()

    @track_usage
    def create_feature(self, org_id: int, featureset_id: int, name: str, code: str, description: str, column_id: int, status: str, tags: List[str]):
        feature = api.FeatureCreate(name=name,
                                    code=code,
                                    description=description,
                                    featureSetId=featureset_id,
                                    columnId=column_id,
                                    organizationId=org_id,
                                    orchestrationStatus=status,
                                    tags=tags)
        return self._post("/features", feature.dict(exclude_unset=True), api_version=1).json()

    @track_usage
    def update_feature(self, feature_id: int, org_id: int, featureset_id: int, name: str, code: str, description: str, column_id: int, status: str, tags: List[str]):
        feature = api.FeatureUpdate(id=feature_id,
                                    name=name,
                                    code=code,
                                    description=description)
        #TODO: We won't be able to update these fields until we create a V1 patch features endpoint
                                    #featureSetId=featureset_id,
                                    #columnId=column_id,
                                    #organizationId=org_id,
                                    #orchestrationStatus=status,
                                    #tags=tags)
        return self._patch(f"features/{feature_id}", feature.dict()).json()

    @track_usage
    def create_feature_set(self, name, datasource_id, table_name, granularity=None):
        feature_set = api.FeatureSetCreate(name=name,
                                           snowflakeTable=table_name,
                                           dataSource=api.DataSourceBase(id=datasource_id),
                                           granularity=granularity)
        return self._post("feature-sets", feature_set.dict()).json()

    @track_usage
    def update_feature_set(self, feature_set_id, name, datasource_id, table_name, granularity=None):
        feature_set = api.FeatureSetUpdate(id=feature_set_id,
                                           name=name,
                                           snowflakeTable=table_name,
                                           dataSource=api.DataSourceBase(id=datasource_id),
                                           granularity=granularity)
        return self._patch(f"feature-sets/{feature_set_id}", feature_set.dict()).json()

    @track_usage
    def publish_feature_set(self, name, datasource_id, table_name, granularity=None):
        '''
        Creates or updates a featureset depending on existence of the defined parameters
        '''
        fs = self._get_item("feature-sets", { "snowflakeTable": table_name})
        if fs:
            feature_set = self.update_feature_set(fs['id'], name, datasource_id, table_name, granularity)
        else:
            feature_set = self.create_feature_set(name, datasource_id, table_name, granularity)
        return feature_set

    def dimensionality_name(self, datatype, granularity):
        return "{} - {}".format(datatype, str(granularity).title())

    @track_usage
    def publish_dimensionality(self, org_id, dimension_type, granularity):
        """
        Creates or returns a dimensionality depending on existence of the defined parameters

        Dimensionality is a named pairing of a datatype and a granularity. Note in some cases the
        granularity is actually a data type.
        """
        if dimension_type is None:
            if granularity.lower() in ['hour', 'day', 'week', 'month', 'quarter', 'year']:
                dimension_type = 'DateTime'
            else:
                dimension_type = "Custom"
        elif dimension_type.lower() == "datetime":
            dimension_type = "DateTime"
        else:
            dimension_type = dimension_type.title()
        dimensionality_name = self.dimensionality_name(dimension_type, granularity)

        # Check for a 'dimensionality' record that corresponds to the the dimensions
        # datatype and granularity.
        dimensionality = self.get_dimensionality(dimensionality_name)
        if dimensionality:
            return dimensionality
        else:
            return self.create_dimensionality(org_id, dimensionality_name, dimension_type, granularity)

    @track_usage
    def get_column(self, name, featureset_id):
        '''
        Returns a column matching a name in the specified featureset
        '''
        items = self._get_items("rcolumns", {"name": name})
        for i in items:
            if featureset_id == i['featureSet']['id']:
                return i
        return None

    @track_usage
    def publish_dimension(self, org_id, featureset_id, name, data_type, dimension_type=None, granularity=None):
        '''
        Creates or updates a dimension depending on existence of the defined parameters
        '''
        dimemnsionality = self.publish_dimensionality(org_id, dimension_type, granularity)
        dimemnsionality_id = dimemnsionality['id']
        dim = self.get_column(name, featureset_id)
        if dim:
            dimension = self.update_column(dim['id'], name, data_type, featureset_id, dimemnsionality_id)
        else:
            dimension = self.create_column(name, data_type, featureset_id, dimemnsionality_id)
        return dimension

    @track_usage
    def publish_feature(self, org_id, featureset_id, name, data_type, code=None, description=None, granularity=None, status=None, tags=None):
        '''
        Creates or updates a feature depending on existence of the defined parameters
        '''
        if code is None:
            code = name
        if description is None:
            description = f"Feature that contains {name} data"
        if status is None:
            status = 'Sandboxed'  
        if granularity is None:
            dimension_id = None    
        else:
            dimemnsionality = self.publish_dimensionality(org_id, None, granularity)
            dimension_id = dimemnsionality['id']
        
        ft = self._get_item("features", {"code": code})
        if ft:
            column = self.update_column(ft['column']['id'], code, data_type, featureset_id, dimension_id)
            feature = self.update_feature(ft['id'], org_id, featureset_id, name, code, description, ft['column']['id'], status, tags)
        else:
            column = self.create_column(code, data_type, featureset_id, dimension_id)
            feature = self.create_feature(org_id, featureset_id, name, code, description, column['id'], status, tags)
        return feature

    @track_usage
    def publish_datasource(self, org_id, ds_name):
        '''
        Creates or returns a datasource depending on of the defined parameters
        '''
        # Check for a 'dimensionality' record that corresponds to the the dimensions
        # datatype and granularity.
        ds = self.get_datasource(ds_name)
        if ds:
            return ds
        else:
            return self.create_datasource(org_id, ds_name)

    def _confirm_df_columns(self, dataframe, dimensions, features):
        df_columns = list(dataframe.columns)
        missing_dims = []
        missing_features = []
        for dim in dimensions:
            if dim not in df_columns:
                missing_dims.append(dim)
        for ft in features:
            if ft not in df_columns:
                missing_features.append(ft)
        if missing_dims or missing_features:
            raise Exception(f"Specified columns do not exist in dataframe: "
                            f"Dimensions({missing_dims}) Features({missing_features})")

    def _generate_featureset_name(self, dimensions, timestamp):
        return f"pandas_by_{'-'.join(dimensions)}_{timestamp}"

    def _snowflakify_name(self, name):
        '''
        param name: string
        return: string
        Converts a string to a snowflake compliant value
        Removes double quotes, replaces dashes with underscores, casts to upper case
        '''
        return name.replace("-", "_").replace('"', '').upper()

    def _snowflakify_list(self, list_in):
        '''
        param list_in: list
        return list_out: list
        Changes a list of columns to Snowflake compliant names
        '''
        list_out = [self._snowflakify_name(n) for n in list_in]
        return list_out

    def _snowflakify_dataframe(self, dataframe: pd.DataFrame):
        '''
        param dataframe: dataframe holding columns
        param column: list of column names that need to change
        Renames all columns in a pandas dataframe to Snowflake compliant names
        '''
        schema = self._schema_for_dataframe(dataframe)
        cols = {}
        for r in schema:
            oldc = r
            newc = self._snowflakify_name(oldc)
            cols[oldc] = newc
        dataframe.rename(columns=cols, inplace=True)

    def _dataframe_to_snowflake(self, dataframe, table_name):
        # TODO: Move this function to a lambda
        # we want to grant a user access to create tables in the schema
        # currently we're doing it in the model_training lambda
        # I can't be done in this method due to role limitations at runtime
        conn, creds = self._snowflake_connection(self.get_member())
        with conn.cursor() as cur:
            # Create the table in Snowflake
            tablesql = self._generate_ddl_from_dataframe(dataframe, table_name)
            cur.execute(tablesql)
        # load data from df to SF table
        write_pandas(conn, dataframe, table_name)

    def _generate_ddl_from_dataframe(self, df, table_name):
        sql_text = pd.io.sql.get_schema(df.reset_index(), table_name)
        sql_text = sql_text.replace('CREATE TABLE', 'CREATE OR REPLACE TABLE')
        sql_text = sql_text.replace('"', '')
        return sql_text

    def _schema_for_dataframe(self, df):
        from pandas.io.json import build_table_schema
        schema_list = build_table_schema(df)
        schema = {}
        for column in schema_list['fields']:
            name = column['name']
            schema[name] = column
        return schema

    @track_usage
    def publish_features_from_df(self, dataframe: pd.DataFrame, dimensions: List[str], features: List[str], granularity: str=None, tags: List[str]=None):
        '''
        Creates a featureset from a pandas dataframe

        :dataframe: Pandas DataFrame containing all columns that will be registered with Rasgo
        :param dimensions: List of columns in df that should be used for joins to other featursets
        :param features: List of columns in df that should be registered as features in Rasgo
        :param granularity: Datetime grain to be added to all features in the df
        :param tags: List of tags to be added to all features in the df
        :return: description of the featureset created
        '''
        # todo: Optionally generate list of feature columns from the dataframe columns, ie - all non-dimensions are features
        # todo: Add option to specify a featureset name + add check that it exists.
        #Type checking
        if type(dimensions) != list:
            raise TypeError('Dimensions must be passed in as a list')
        if type(features) != list:
            raise TypeError('Features must be passed in as a list')
        if tags:
            if type(tags) != list:
                raise TypeError('Tags must be passed in as a list')
        
        member = self.get_member()
        org_id = member.organization_id()
        now = datetime.now()
        timestamp = now.strftime("%Y_%m_%d_%H_%M")
        datasource = self.publish_datasource(org_id, 'PANDAS')

        # Convert all strings to work with Snowflake
        dimensions = self._snowflakify_list(dimensions)
        features = self._snowflakify_list(features)
        self._snowflakify_dataframe(dataframe)

        # Confirm each named dimension and feature exists in the dataframe.
        self._confirm_df_columns(dataframe, dimensions, features)

        # Generate featureset name.
        featureset_name = self._generate_featureset_name(dimensions, timestamp)

        # Create a table in Snowflake with the subset of columns we're interested in, name table after featureset.
        all_columns = dimensions + features
        exportable_df = dataframe[all_columns]
        table_name = self._snowflakify_name(featureset_name)
        self._dataframe_to_snowflake(exportable_df, table_name)

        # Add a reference to the FeatureSet
        featureset_name = table_name
        featureset = self.create_feature_set(featureset_name, datasource['id'], table_name, granularity)
        schema = self._schema_for_dataframe(dataframe)

        return_featureset = {}
        return_featureset['id'] = featureset['id']
        return_featureset['name'] = featureset['name']
        return_featureset['granularity'] = featureset['granularity']
        return_featureset['dataSource'] = featureset['dataSource']['name']
        return_featureset['organization'] = featureset['dataSource']['organization']['name']
        return_featureset['snowflakeTable'] = featureset['snowflakeTable']

        # Add references to all the dimensions
        return_dimensions = {}
        for d in dimensions:
            column = schema[d]
            data_type = column['type']
            dimension_name = column['name']
            dimension = self.publish_dimension(org_id, featureset['id'], dimension_name, data_type, None, granularity)
            return_dimensions['name'] = dimension['name']
        return_featureset['dimensions'] = return_dimensions

        # Add references to all the features
        return_features = {}
        for f in features:
            column = schema[f]
            data_type = column['type']
            code = column['name']
            feature_name = f"PANDAS_{code}_{timestamp}"
            status = 'Sandboxed'
            tags.append('Pandas')
            feature = self.publish_feature(org_id, featureset['id'], feature_name, data_type, code, None, granularity, status, tags)
            return_features['id'] = feature['id']
            return_features['name'] = feature['name']
            return_features['column'] = feature['code']
        return_featureset['features'] = return_features
        return return_featureset

    @track_usage
    def publish_features_from_yml(self, yml_file, orchestrationStatus='Sandboxed'):
        '''
        Creates a featureset from a yml file

        :param yml_file: Rasgo compliant yml file that describes the featureset(s) being created
        :return: description of the featureset created
        '''
        #TODO: add update logic in publish methods
        with open(yml_file) as fobj:
            featuresets = yaml.load(fobj, Loader = yaml.SafeLoader)
        for fs in featuresets:
            #publish featureset
            member = self.get_member()
            org_id = member.organization_id()
            datasource = self.publish_datasource(org_id, fs["datasource"])            
            featureset_name = fs.get("name", fs["table"])
            snowflake_table = fs["table"]
            if not snowflake_table:
                raise Exception("A valid table name is required")
            granularity = fs.get('granularity', None)
            featureset = self.publish_feature_set(featureset_name, datasource['id'], snowflake_table, granularity)

            return_featureset = {}
            return_featureset['id'] = featureset['id']
            return_featureset['name'] = featureset['name']
            return_featureset['granularity'] = featureset['granularity']
            return_featureset['dataSource'] = featureset['dataSource']['name']
            return_featureset['organization'] = featureset['dataSource']['organization']['name']
            return_featureset['snowflakeTable'] = featureset['snowflakeTable']

            #publish dimensions
            return_dimensions = {}
            for dim in fs['dimensions']:
                name = dim.get('name')
                data_type = dim.get('data_type')
                d = self.publish_dimension(org_id, featureset['id'], name, data_type, None, granularity)
                return_dimensions['name'] = d['name']
            return_featureset['dimensions'] = return_dimensions

            #publish features
            return_features = {}
            for feature in fs['features']:
                name = feature['display_name']
                code = feature.get('name', name)
                data_type = feature.get('data_type')
                description = feature.get('description', f"Feature that contains {name} data")
                tags = feature.get('tags')
                f = self.publish_feature(org_id, featureset['id'], name, data_type, code, description, granularity, orchestrationStatus, tags)
                return_features['id'] = f['id']
                return_features['name'] = f['name']
                return_features['column'] = f['code']
            return_featureset['features'] = return_features
        return featureset

    @staticmethod
    def _snowflake_connection(member) -> (SnowflakeConnection, dict):
        """
        Constructs connection object for Snowflake data platform
        :param member: credentials for Snowflake data platform

        :return: Connection object to use for query execution
        """
        creds = member.snowflake_creds()
        conn = snowflake.connect(**creds)
        return conn, creds

    @staticmethod
    def _make_select_statement(table_metadata, filters, limit) -> tuple:
        """
        Constructs select * query for table
        """
        query = "SELECT * FROM {database}.{schema}.{table}".format(**table_metadata)
        values = []
        if filters:
            comparisons = []
            for k, v in filters.items():
                if isinstance(v, list):
                    comparisons.append(f"{k} IN ({', '.join(['%s'] * len(v))})")
                    values += v
                else:
                    comparisons.append(f"{k}=%s")
                    values.append(v)
            query += " WHERE " + " and ".join(comparisons)
        if limit:
            query += " LIMIT {}".format(limit)
        return query, values

    @staticmethod
    @track_usage
    def _run_query(conn, query: str, params):
        """
        Execute a query on the [cloud] data platform.

        :param conn: TODO -> abstract the cloud data platform connection
        :param query: String to be executed on the data platform
        :return:
        """
        return conn.cursor().execute(query, params)

    @deprecated("This function has been deprecated, use `get_models` instead.")
    @track_usage
    def get_lists(self) -> List[Model]:
        """
        Deprecated function.  Renamed to `get_models.`
        """
        return self.get_models()

    @deprecated("This function has been deprecated, use `get_model` instead.")
    @track_usage
    def get_feature_list(self, list_id) -> Model:
        """
        Deprecated function.  Renamed to `get_model.`
        """
        return self.get_model(model_id=list_id)



import os
import json
import logging
import requests
import functools

from pyrasgo.version import __version__ as pyrasgo_version

PRODUCTION_HEAP_KEY = '540300130'
STAGING_HEAP_KEY = '3353132567'

PRODUCTION = "PRODUCTION"
STAGING = "STAGING"
LOCAL = "LOCAL"
ENDPOINTS = {
    PRODUCTION: "api.rasgoml.com",
    LOCAL: "localhost"
}

HEAP_URL = "https://heapanalytics.com/api"


# props_api = f"{base_url}/add_user_properties"


def track_usage(func):
    @functools.wraps(func)
    def decorated(self, *args, **kwargs):
        from pyrasgo.connection import Connection
        if issubclass(self.__class__, Connection) or isinstance(self, Connection):
            self: Connection
            if self.environment == LOCAL:
                logging.info(f"Called {func.__name__} with parameters: {kwargs}")
            else:
                try:
                    username = self._username
                except AttributeError:
                    username = self._get_profile().get("username", "Unknown"),

                try:
                    track_call(
                        app_id=PRODUCTION_HEAP_KEY if self._hostname == ENDPOINTS[PRODUCTION] else STAGING_HEAP_KEY,
                        user_id=self._user_id,
                        event=func.__name__,
                        properties={"hostname": os.environ.get('RASGO_DOMAIN', ENDPOINTS[PRODUCTION]),
                                    "source": "pyrasgo",
                                    "class": self.__class__.__name__,
                                    "version": pyrasgo_version,
                                    "username": username,
                                    "input": args,
                                    **kwargs})
                except Exception:
                    logging.debug(f"Called {func.__name__} with parameters: {kwargs}")
        else:
            logging.debug(f"Cannot track functions called from {self.__class__.__name__} class.")
        return func(self, *args, **kwargs)
    return decorated


def track_call(app_id: str,
               user_id: int,
               event: str,
               properties: dict = None):
    """
    Send a "track" event to the Heap Analytics API server.

    :param event: event name
    :param properties: optional, additional event properties
    """
    data = {"app_id": app_id,
            "identity": user_id,
            "event": event}

    if properties is not None:
        data["properties"] = properties

    response = requests.post(url=f"{HEAP_URL}/track",
                             data=json.dumps(data),
                             headers={"Content-Type": "application/json"})
    response.raise_for_status()
    return response



from typing import List
from collections import Sequence

from pyrasgo.connection import Connection
from pyrasgo.enums import Granularity
from pyrasgo.monitoring import track_usage
from pyrasgo.namespace import Namespace
from pyrasgo.schemas import feature as api


class FeatureList(Sequence):
    """
        Convenience class to enable simpler console presentation,
        iteration and searching through lists of Features objects
    """

    def __init__(self, api_object, api_key=None, user_id=None, username=None, org_id=None):
        # TODO: Would this be more useful as a set not a sequence?
        self._api_key = api_key
        self.data = sorted([Feature(api_object=entry, api_key=api_key,
                                    user_id=user_id, username=username,
                                    org_id=org_id) for entry in api_object], key=lambda x: x.id)

    def __getitem__(self, i: int):
        return self.data[i]

    def __len__(self) -> int:
        return len(self.data)

    def __str__(self):
        ids = [str(feature.id) for feature in self]
        return (f"Features({len(self.data)} total, "
                f"ids: [{','.join(ids if len(self) < 7 else ids[0:3] + ['...'] + ids[-3:])}])")

    def __add__(self, other):
        # TODO: Should this be a set operations??
        if isinstance(other, Feature):
            return type(self)([feature.to_dict() for feature in self.data + [other]], api_key=self._api_key)
        if isinstance(other, type(self)):
            return type(self)([feature.to_dict() for feature in self.data + other.data], api_key=self._api_key)
        if isinstance(other, list) and all(
                [isinstance(entry, Namespace) or isinstance(entry, Feature) for entry in other]):
            return type(self)([feature.to_dict() for feature in self.data + other], api_key=self._api_key)
        raise TypeError(f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'")

    def __repr__(self):
        return str(self)

    def filter(self, **kwargs):
        return [feature for feature in self.data
                if [feature.__getattr__(key) for key in kwargs.keys()] == list(kwargs.values())]


class Feature(Connection):
    """
    Stores the properties for a feature
    """

    def __init__(self, api_object, **kwargs):
        # TODO: Determine pathway to create feature via command line? Do we want to allow this?
        super().__init__(**kwargs)
        self._feature = api.Feature(**api_object)

    def __getattr__(self, item):
        try:
            return self._feature.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Feature(id={self.id}, name={self.name}, description={self.description})"

    @track_usage
    def get(self):
        """
        Updates the Feature object's attributes from the API
        """
        self._feature = api.Feature(**self._get(f"/features/{self.id}", api_version=1).json())

    @track_usage
    def add_tag(self, tag: str):
        if tag in self._feature.tags:
            return
        self._feature = api.Feature(**self._put(f"/features/{self.id}/tags",
                                                api_version=1, _json={"tags": [tag, *self._feature.tags]}).json())

    @track_usage
    def add_tags(self, tags: List[str]):
        if all(tag in self._feature.tags for tag in tags):
            return
        self._feature = api.Feature(**self._put(f"/features/{self.id}/tags",
                                                api_version=1, _json={"tags": [*tags, *self._feature.tags]}).json())

    @track_usage
    def delete_tag(self, tag: str):
        if tag not in self._feature.tags:
            return
        self._feature = api.Feature(**self._delete(f"/features/{self.id}/tags",
                                                   api_version=1, _json={"tags": [tag]}).json())

    @track_usage
    def delete_tags(self, tags: List[str]):
        if all(tag not in self._feature.tags for tag in tags):
            return
        self._feature = api.Feature(**self._delete(f"/features/{self.id}/tags",
                                                   api_version=1, _json={"tags": [*tags]}).json())

    @property
    @track_usage
    def datasource(self):
        """
        Retrieves datasource name for the feature
        """
        # TODO: This is tightly coupled to the API's response object
        data_source = self._feature.dataSource
        if data_source:
            return data_source.name
        return None

    @property
    @track_usage
    def granularities(self):
        """
        Retrieves granularities for the feature
        """
        # TODO: This is tightly coupled to the API's response object
        granularities = self._feature.granularities
        if granularities:
            return [Granularity(granularity.name) for granularity in granularities]
        return []

    @property
    @track_usage
    def featureSet(self):
        """
        Retrieves the feature set for the feature
        """
        # TODO: This is tightly coupled to the API's response object
        feature_set = self._feature.featureSet
        if feature_set:
            return feature_set.name
        return None

    @property
    @track_usage
    def tags(self) -> List[str]:
        return self._feature.tags



from typing import Union, List

import time
from tqdm import tqdm

from pyrasgo.connection import Connection
from pyrasgo.enums import ModelType
from pyrasgo.feature import Feature, FeatureList
from pyrasgo.member import Member
from pyrasgo.monitoring import track_usage
from pyrasgo.namespace import Namespace


class Model(Connection):
    """
    Stores all the features for a model
    """

    def __init__(self, api_object, **kwargs):
        self._namespace = Namespace(**api_object)
        super().__init__(**kwargs)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Model(id={self.id}, name={self.name})"

    def __getattr__(self, item):
        try:
            return self._namespace.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

    @property
    @track_usage
    def type(self):
        """
        Returns the model type of the model
        """
        return ModelType(self._namespace.type)

    @property
    @track_usage
    def features(self):
        """
        Returns the features within the model
        """
        return FeatureList([feature.to_dict() for feature in self._namespace.features], api_key=self._api_key)

    @features.setter
    def features(self, features: Union[List[Feature], FeatureList, Feature]):
        # Update the namespace
        self._namespace.__setattr__('features',
                                    [Namespace(**feature.to_dict()) for feature in self.features + features])

    def get(self):
        """
        Updates the Model object's attributes from the API
        """
        self._namespace = Namespace(**self._get(f"/models/{self.id}").json())

    def get_author(self):
        """
        Returns the full author object of the model (including credentials)
        """
        return Member(self.author.to_dict())

    @track_usage
    def add_feature(self, feature: Feature) -> None:
        """
        Adds a single feature to the model
        """
        self.features = feature
        self._patch(f"/models/{self.id}/features", api_version=1,
                    _json={"featureIds": [str(feature.id)]})

    @track_usage
    def add_features(self, features: Union[List[Feature], FeatureList]) -> None:
        """
        Adds a FeatureList or a list of Features to the model.
        """
        self.features = features
        self._patch(f"/models/{self.id}/features", api_version=1,
                    _json={"featureIds": [str(feature.id) for feature in features]})

    @track_usage
    def is_data_ready(self) -> bool:
        """
        Performs check against API for training data readiness, if true, a dataframe can be pulled down.
        :return:
        """
        r = self._get("/trainings/latest")
        if any([model['state'] == 'done' for model in r.json() if model['model']['id'] == self.id]):
            return True
        if any([model['state'] == 'error' for model in r.json() if model['model']['id'] == self.id]):
            raise SystemError("There's been an issue with the training data generation")
        return False

    @track_usage
    def generate_training_data(self) -> None:
        """
        Triggers the generation of the model's training data.
        """
        self._post(f"/trainings/", _json={"modelId": int(self.id), "userId": self._user_id}, api_version=0)
        for _ in tqdm(range(0, 10)):
            if self.is_data_ready():
                break
            time.sleep(1)

    def snowflake_table_metadata(self, creds):
        metadata = {
            "database": creds.get("database", "rasgoalpha"),
            "schema": creds.get("schema", "public"),
            "table": self._snowflake_table_name(),
        }
        return metadata

    def _snowflake_table_name(self):
        table_name = self.dataTableName
        if table_name is None:
            raise ValueError("No table found for model '{}'".format(self.id))
        return table_name



import os
import requests

from urllib.parse import urlencode, quote
from requests.exceptions import HTTPError

PRODUCTION = "PRODUCTION"
STAGING = "STAGING"
LOCAL = "LOCAL"
ENDPOINTS = {
    PRODUCTION: "api.rasgoml.com",
    STAGING: "staging-rasgo-proxy.herokuapp.com",
}


class InvalidApiKeyException(Exception):
    pass


class Connection(object):
    """
    Base class for all Rasgo objects to facilitate API calls.
    TODO: Possibly move to dependency injection of class with pinject or dependency-injector.
    """

    def __init__(self, api_key=None, user_id=None, username=None, org_id=None):
        self._api_key = api_key

        rasgo_domain = os.environ.get("RASGO_DOMAIN", ENDPOINTS[PRODUCTION])
        if rasgo_domain == ENDPOINTS[PRODUCTION]:
            self.environment = PRODUCTION
        elif rasgo_domain == ENDPOINTS[STAGING]:
            self.environment = STAGING
        else:
            self.environment = LOCAL
        self._hostname = rasgo_domain

        if self.__class__.__name__ == 'Connection':
            try:
                user_profile = self._get_profile()
            except HTTPError:
                raise InvalidApiKeyException(f"The API key provided ({api_key[:5]}...{api_key[-5:]}) is not valid.")
            self._user_id = user_profile.get('id')
            self._username = user_profile.get('username')
            self._org_id = user_profile.get('organizationId')
        else:
            if not api_key:
                raise ValueError("The API key must be provided to the constructor if not built by Connection.")
            self._user_id = user_id
            self._username = username
            self._org_id = org_id

    def _get_profile(self):
        return self._get('/users/me', api_version=1).json()

    def _url(self, resource, api_version=None):
        if '/' == resource[0]:
            resource = resource[1:]
        return f"https://{self._hostname}/{'' if api_version is None else f'v{api_version}/'}{resource}"

    def find(self, resource, equality_filters):
        search_strings = []
        for k, v in equality_filters.items():
            filterstr = k + "||$eq||" + v
            search_strings.append(filterstr)
        params = {
            "filter": search_strings
        }
        params = urlencode(params, quote_via=quote, doseq=True)
        response = self._get(resource, params)
        if response.status_code == 404:
            return None
        else:
            response.raise_for_status()
            return response.json()

    def _delete(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs DELETE request to Rasgo API as defined within the class instance.

        :param resource: Target resource to DELETE from API.
        :param _json: JSON object to send in DELETE request
        :param params: Additional parameters to specify for POST request.
        :return: Response object containing content returned.
        """
        response = requests.delete(self._url(resource, api_version),
                                json=_json,
                                headers=self._headers(self._api_key),
                                params=params or {})
        response.raise_for_status()
        return response

    def _get(self, resource, params=None, api_version=None) -> requests.Response:
        """
        Performs GET request to Rasgo API as defined within the class instance.

        :param endpoint: Target resource to GET from API.
        :param params: Additional parameters to specify for GET request.
        :return: Response object containing content returned.
        """
        response = requests.get(self._url(resource, api_version),
                                headers=self._headers(self._api_key),
                                params=params or {})
        response.raise_for_status()
        return response

    def _patch(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs PATCH request to Rasgo API as defined within the class instance.

        :param resource: Target resource to PATCH from API.
        :param _json: JSON object to send in PATCH request
        :param params: Additional parameters to specify for PATCH request.
        :return: Response object containing content returned.
        """
        response = requests.patch(self._url(resource, api_version),
                                  json=_json,
                                  headers=self._headers(self._api_key),
                                  params=params or {})
        response.raise_for_status()
        return response

    def _post(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs POST request to Rasgo API as defined within the class instance.

        :param resource: Target resource to POST from API.
        :param _json: JSON object to send in POST request
        :param params: Additional parameters to specify for POST request.
        :return: Response object containing content returned.
        """
        response = requests.post(self._url(resource, api_version),
                                 json=_json,
                                 headers=self._headers(self._api_key),
                                 params=params or {})
        response.raise_for_status()
        return response

    def _put(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs PUT request to Rasgo API as defined within the class instance.

        :param resource: Target resource to PUT from API.
        :param _json: JSON object to send in PUT request
        :param params: Additional parameters to specify for PUT request.
        :return: Response object containing content returned.
        """
        response = requests.put(self._url(resource, api_version),
                                json=_json,
                                headers=self._headers(self._api_key),
                                params=params or {})
        response.raise_for_status()
        return response

    def _get_items(self, resource, equality_filters):
        search_strings = []
        for k, v in equality_filters.items():
            filterstr = k + "||$eq||" + v
            search_strings.append(filterstr)
        params = {
            "filter": search_strings
        }
        params = urlencode(params, quote_via=quote, doseq=True)
        try:
            response = self._get(resource, params)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return None
            else:
                raise e
        if response.status_code == 404:
            return None
        else:
            response.raise_for_status()
            return response.json()

    def _get_item(self, resource, equality_filters, allow_multiple=False):
        results = self._get_items(resource, equality_filters)
        if results is None:
            return None
        elif 1 < len(results):
            if allow_multiple:
                return results[-1]
            else:
                raise ValueError("Multiple returned for {}, check your filters".format(resource))
        elif 0 == len(results):
            return None
        else:
            return results[0]

    @staticmethod
    def _headers(api_key) -> dict:
        if not api_key:
            raise ValueError("Must provide an API key to access the endpoint")
        return {
            "accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        }



from pydantic import BaseModel

from pyrasgo.schemas.organization import Organization


class Dimensionality(BaseModel):
    id: int


class DimensionalityCreate(BaseModel):
    name: str
    dimension_type: str
    granularity: str
    organization: Organization



from enum import Enum
from typing import Union


class Granularity(Enum):
    """
    Enum class to describe allowed types of granularities
    """
    # TODO: This exists to support validation prior to posting from the API.
    #       For ease of maintenance, this should be handled through API returns.
    HOUR = 'hour'
    DAY = 'day'
    WEEK = 'week'
    MONTH = 'month'
    YEAR = 'year'
    QUARTER = 'quarter'
    FIPS = 'FIPS'
    MOVIE = 'movie'
    CBG = 'CBG'
    TICKER = 'ticker'
    DMA = 'DMA'
    TERM = 'term'
    TEAM = 'team'
    AIRLINE = 'airline'
    CUSTOM = 'custom'


class ModelTypes(Enum):
    """
    Enum class to describe allowed types of models
    """
    # TODO: This exists to support validation prior to posting from the API.
    #       For ease of maintenance, this should be handled through API returns.
    TIMESERIES = 'time series'


class ModelType(object):
    """
    Class to wrap the ModelTypes enum for additional functionality

    (Extending the enum class is not allowed, and subclassing is possible, but unnecessary)
    """

    def __init__(self, name: str):
        self._type = ModelTypes(name)
        # TODO: Lookup compatible when other model types and/or granularities are supported.
        self._compatible_granularities = [Granularity.DAY, Granularity.MONTH,
                                          Granularity.YEAR, Granularity.WEEK, Granularity.QUARTER]

    @property
    def name(self):
        return self._type.name

    @property
    def value(self):
        return self._type.value

    @property
    def compatible_granularities(self):
        return self._compatible_granularities

    def is_compatible(self, granularity: Union[str, Granularity]) -> bool:
        # If not enum, convert to enum first.
        try:
            granularity.name
        except AttributeError:
            granularity = Granularity(granularity)
        return granularity in self._compatible_granularities



from .column import Column, ColumnCreate, ColumnUpdate
from .data_source import DataSourceBase, DataSourceCreate
from .dimensionality import DimensionalityCreate, Dimensionality
from .feature import FeatureCreate, FeatureUpdate
from .feature_set import FeatureSetBase, FeatureSetCreate, FeatureSetUpdate
from .organization import Organization
